//*=================================================================================*
// Inertia Mesh Viewer for Windows
//*=================================================================================*

//*=================================================================================*
// Headers
//*=================================================================================*
#include <windows.h>
#include "dibsect.h"            // Dude's DIB lib
#include "..\inertia\types.h"   // Inertia types
#include "..\inertia\inertia.h" // Inertia functions
#include "trkball.h"            // 2D Trackball interface
#include "resource.h"			// resource file stuff

//*=================================================================================*
// Global variables
//*=================================================================================*
ViewObject    View;                  // a view for which to see everything
VectorObject  Obj;                   // the object we are going to look at
CDIBSection8 *VirtualScreen;         // 8-bit Device Independant Bitmap

float         ZScale;                // z step of the object (for mouse control)  
float         ZLoc;                  // Z location of the object

int           TransFlag   = 0;       // start without transparency
int           ShadeFlag   = LAMBERT; // start with lambert shading
int           SurfaceFlag = SMOOTH;  // start with smooth surface model
int           RenderEvent = 0;       // recording var for when key app events occur

char *ErrorCaption = "Inertia Mesh Viewer Error"; // Caption for error dialog/message boxes

//*=================================================================================*
// Function headers
//*=================================================================================*
LRESULT CALLBACK EventHandler(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK DialogEventHandler(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
int  GeneralInit();
int  CreateVirtualPage(HWND hwnd);
void CleanUp();
void RunApplication();
void LoadNewGVOFile(HWND hwnd);
void PaintScreen();
void MouseAction(int Button, POINTS MouseCoord);
void UpdateSurfaceChecks(HWND hwnd, int NewSurface);
void UpdateShadeChecks(HWND hwnd, int NewShade);
void UpdateTransparency(HWND hwnd);

//*=================================================================================*
// MAIN
//*=================================================================================*
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR tmp3, int nCmdShow)
{
	MSG msg;
	HWND hwnd;
	WNDCLASS wndClass;

	if (!GeneralInit())  // start off by initializing some Inertia things
	{
		return(FALSE);
	}

	if (hPrevInstance == NULL)  // Create a new window class if necessairy
	{
		memset(&wndClass, 0, sizeof(wndClass));
		wndClass.style         = CS_HREDRAW | CS_VREDRAW;
		wndClass.lpfnWndProc   = EventHandler;        // use our event handler
		wndClass.hInstance     = hInstance;
		wndClass.hCursor       = LoadCursor(NULL, IDC_ARROW);
		wndClass.hIcon         = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON1));
		wndClass.lpszClassName = "Inertia_Window_Class";
		wndClass.lpszMenuName  = MAKEINTRESOURCE(IDR_MENU1);

		if (!RegisterClass(&wndClass))
			return FALSE;
	}

	// Create/Open the window
	hwnd = CreateWindow("Inertia_Window_Class", "Inertia Mesh Viewer",
						WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, 0, 640, 480, NULL, NULL,
						hInstance, NULL);

	// Create the virtual page for us to render on
	if (!CreateVirtualPage(hwnd))
	{
		CleanUp();
		return(FALSE);
	}

	ShowWindow(hwnd, nCmdShow);
	UpdateWindow(hwnd);

	// main execution loop
	// This loop isn't the best way to do things
	// ideally we would set up a timer messaging system
	// to handle the animation. The current way is more
	// of a CPU hog.
	while (TRUE)  
	{
		RunApplication();  // run the Inertia portion of the application

		// Check for any windows messages
        if (PeekMessage (&msg, NULL, 0, 0, PM_REMOVE))
		{
			if ((msg.message == WM_QUIT) || (msg.message == WM_DESTROY))
			{
			  CleanUp();
			  return (msg.wParam);
			}

          TranslateMessage (&msg);
          DispatchMessage (&msg);
		}
	}

	return(msg.wParam);
}

//*=================================================================================*
// RunApplication - main application loop where everything is rendered
//*=================================================================================*
void RunApplication()
{
	if (RenderEvent)  // only update the screen when it changes
	{
		PaintScreen();
		RenderEvent = 0;
	}
}

//*=================================================================================*
// EventHandler - Handles windows specific events
//*=================================================================================*
LRESULT CALLBACK EventHandler(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_MOUSEMOVE: MouseAction(wParam, MAKEPOINTS(lParam)); break;
	case WM_PAINT: PaintScreen(); break;
	case WM_COMMAND:
		{
			RenderEvent = 1;  // a rendering even has occured    
			switch (wParam & 0xffff) 
			{
			case (ID_INER_ABOUT & 0xffff):        DialogBox((HINSTANCE)GetWindowLong(hwnd, GWL_HINSTANCE), MAKEINTRESOURCE(IDD_DIALOG1), hwnd, (DLGPROC)DialogEventHandler); break;
			case (ID_INER_FILE_QUIT & 0xffff):    CleanUp(); PostQuitMessage(0); exit(0); break;
			case (ID_INER_FILE_OPEN & 0xffff):    LoadNewGVOFile(hwnd); break;
			case (ID_INER_SMOOTH & 0xffff):       UpdateSurfaceChecks(hwnd, SMOOTH); break;
			case (ID_INER_TEXTURE & 0xffff):      UpdateSurfaceChecks(hwnd, TEXTURE); break;
			case (ID_INER_PERSPTEXTURE & 0xffff): UpdateSurfaceChecks(hwnd, PERSPTEXTURE); break;
			case (ID_INER_REFLECTION & 0xffff):   UpdateSurfaceChecks(hwnd, REFLECTION); break;
			case (ID_INER_UNSHADED & 0xffff):     UpdateShadeChecks(hwnd, UNSHADED); break;
			case (ID_INER_AMBIENT & 0xffff):      UpdateShadeChecks(hwnd, AMBIENT); break;     
			case (ID_INER_LAMBERT & 0xffff):      UpdateShadeChecks(hwnd, LAMBERT); break;     
			case (ID_INER_GOURAUD & 0xffff):      UpdateShadeChecks(hwnd, GOURAUD); break;     
			case (ID_INER_PHONG & 0xffff):        UpdateShadeChecks(hwnd, PHONG); break;     
			case (ID_INER_TRANSPARENT & 0xffff):  UpdateTransparency(hwnd); break;     
			}
		} break;
	case WM_SIZE:   // The window has been re-sized
		RenderEvent = 1; 
		delete VirtualScreen;
		if (!CreateVirtualPage(hwnd))
		{
			CleanUp();
			PostQuitMessage(0);
			exit(0);
		}

		SetViewPort(&View, 0, 0, (float)VirtualScreen->GetWidth() - 1, (float)VirtualScreen->GetHeight() - 1, 1);
		return(0);
		break;
	case WM_DESTROY: CleanUp(); PostQuitMessage(0); exit(0); break;
	}

	return (DefWindowProc(hwnd, uMsg, wParam, lParam));
}

//*=================================================================================*
// DialogEventHandler - event handler for ABOUT dialog box
//*=================================================================================*
LRESULT CALLBACK DialogEventHandler(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG: return (TRUE);
	case WM_COMMAND: if (wParam == IDOK)
					 {
						 EndDialog(hwnd, TRUE);
						 return (TRUE);
					 } break;
	}

	return (FALSE);
}

//*=================================================================================*
// UpdateShadeChecks - updates the little check marks in the shading portion of render menu
//*=================================================================================*
void UpdateShadeChecks(HWND hwnd, int NewShade)
{
	HMENU Menu;

	Menu = GetMenu(hwnd);
	Menu = GetSubMenu(Menu, 1);  // grab the second sub-menu (zero-based "Render")

	// start by un-checking the old surface
	switch (ShadeFlag)
	{
	case UNSHADED: CheckMenuItem(Menu, ID_INER_UNSHADED, MF_UNCHECKED); break;
	case AMBIENT:  CheckMenuItem(Menu, ID_INER_AMBIENT,  MF_UNCHECKED); break;
	case LAMBERT:  CheckMenuItem(Menu, ID_INER_LAMBERT,  MF_UNCHECKED); break;
	case GOURAUD:  CheckMenuItem(Menu, ID_INER_GOURAUD,  MF_UNCHECKED); break;
	case PHONG:    CheckMenuItem(Menu, ID_INER_PHONG,    MF_UNCHECKED); break;
	}

	ShadeFlag = NewShade;  // assign the surface type

	// now let us put a check mark by the new menu choice
	switch (ShadeFlag)
	{
	case UNSHADED: CheckMenuItem(Menu, ID_INER_UNSHADED, MF_CHECKED); break;
	case AMBIENT:  CheckMenuItem(Menu, ID_INER_AMBIENT,  MF_CHECKED); break;
	case LAMBERT:  CheckMenuItem(Menu, ID_INER_LAMBERT,  MF_CHECKED); break;
	case GOURAUD:  CheckMenuItem(Menu, ID_INER_GOURAUD,  MF_CHECKED); break;
	case PHONG:    CheckMenuItem(Menu, ID_INER_PHONG,    MF_CHECKED); break;
	}

	SetRenderMask(&Obj, TransFlag | SurfaceFlag | ShadeFlag);   
}

//*=================================================================================*
// UpdateSurfaceChecks - updates the little check marks in the surface portion of render menu
//*=================================================================================*
void UpdateSurfaceChecks(HWND hwnd, int NewSurface)
{
	HMENU Menu;

	Menu = GetMenu(hwnd);
	Menu = GetSubMenu(Menu, 1);  // grab the second sub-menu (zero-based "Render")

	// start by un-checking the old surface
	switch (SurfaceFlag)
	{
	case SMOOTH:       CheckMenuItem(Menu, ID_INER_SMOOTH, MF_UNCHECKED); break;
	case TEXTURE:      CheckMenuItem(Menu, ID_INER_TEXTURE, MF_UNCHECKED); break;
	case PERSPTEXTURE: CheckMenuItem(Menu, ID_INER_PERSPTEXTURE, MF_UNCHECKED); break;
	case REFLECTION:   CheckMenuItem(Menu, ID_INER_REFLECTION, MF_UNCHECKED); break;
	}

	SurfaceFlag = NewSurface;  // assign the surface type

	// now let us put a check mark by the new menu choice
	switch (SurfaceFlag)
	{
	case SMOOTH:       CheckMenuItem(Menu, ID_INER_SMOOTH, MF_CHECKED); break;
	case TEXTURE:      CheckMenuItem(Menu, ID_INER_TEXTURE, MF_CHECKED); break;
	case PERSPTEXTURE: CheckMenuItem(Menu, ID_INER_PERSPTEXTURE, MF_CHECKED); break;
	case REFLECTION:   CheckMenuItem(Menu, ID_INER_REFLECTION, MF_CHECKED); break;
	}

	SetRenderMask(&Obj, TransFlag | SurfaceFlag | ShadeFlag);  
}

//*=================================================================================*
// UpdateTransparency - updates the transparency check mark in the menu
//*=================================================================================*
void UpdateTransparency(HWND hwnd)
{
	HMENU Menu;

	Menu = GetMenu(hwnd);
	Menu = GetSubMenu(Menu, 1);  // grab the second sub-menu (zero-based "Render")

	if (TransFlag)
	{
		// We are turning off transparency, enable object culling
		CheckMenuItem(Menu, ID_INER_TRANSPARENT, MF_UNCHECKED); 
		TransFlag = 0;
		SetCull(&Obj, 1);
	}
	else
	{
		// We are turning on transparency, disable object culling 
		CheckMenuItem(Menu, ID_INER_TRANSPARENT, MF_CHECKED); 
		TransFlag = TRANSPARENT;
		SetCull(&Obj, 0);
	}

	SetRenderMask(&Obj, TransFlag | SurfaceFlag | ShadeFlag);   
}

//*=================================================================================*
// PaintScreen - performs the screen paint
//*=================================================================================*
void PaintScreen()
{
	// check and make sure our application isn't minimized
	if (VirtualScreen->GetWidth() > 0)  
	{
		AddToRenderList(&View, &Obj);                 
		Render((char *)VirtualScreen->GetBuffer(), VirtualScreen->GetPitch(), 240);
		VirtualScreen->Update();  // Page flip
		memset(VirtualScreen->GetBuffer(), 0, VirtualScreen->GetPitch() * VirtualScreen->GetHeight());
	}
}

//*=================================================================================*
// LoadNewGVOFile - loads a GVO file from disk
//*=================================================================================*
void LoadNewGVOFile(HWND hwnd)
{
	OPENFILENAME fn;
	char TempStr[256];

	memset(TempStr, 0, sizeof(TempStr));       // initialize to a NULL filename
	memset(&fn, 0, sizeof(fn));                // clear the structure
	fn.lStructSize = sizeof(OPENFILENAME); 
	fn.lpstrFilter = "GVO Files\0*.GVO\0\0";   // set the default file type
	fn.lpstrFile   = TempStr;
	fn.nMaxFile    = sizeof(TempStr);
	fn.hwndOwner   = hwnd;

	if (!GetOpenFileName(&fn))    // open a Explorer file dialog
	{
		return;  // user probably pressed CANCEL to exit dialog
	}

	FreeVectorObject(&Obj);  // free the exisiting vector object

	// load a new vector object based on the input file
	if (LoadGVO(&Obj, fn.lpstrFile, 0) != I_OK)
	{
		char ErrorString[128];

		sprintf(ErrorString, "Error! %s could not be loaded\n", fn.lpstrFile);
		MessageBox(NULL, ErrorString, ErrorCaption, MB_OK);
		return;
	}

	SetRenderMask(&Obj, TransFlag | ShadeFlag | SurfaceFlag);

	if (TransFlag)
		SetCull(&Obj, 0);  // turn off culling if the object is transparent

	SetBaseColor(&Obj, 36);
	SetObjectTexture(&Obj, 1);  // set object to the texture at index 1
	SetObjectPhongMap(&Obj, 0); // set object to the phong map at index 0

	SetAbsoluteLocation(&Obj, 0, 0, 3 * Obj.Radius); // set initial location of the object

	ZLoc = Obj.Radius * 3;
	ZScale = Obj.Radius / 30.0f;
}

//*=================================================================================*
// MouseAction - performs the mouse action (rotation and zooming)
//*=================================================================================*
void MouseAction(int Button, POINTS MouseCoord)
{
	static int OldX;
	static int OldY;

	if (Button == MK_LBUTTON)  // check if the left button is down
	{
		RenderEvent = 1;
		SetObjectMatrix(&Obj, (float)(MouseCoord.x - OldX) / 256.0f, -(float)(MouseCoord.y - OldY) / 256.0f, 0);
	}

	if (Button == MK_RBUTTON)  // check if the right button is down
	{
		RenderEvent = 1;
		ZLoc += (float)(MouseCoord.y - OldY) * ZScale;
		SetAbsoluteLocation(&Obj, 0, 0, ZLoc);
	}

	OldX = MouseCoord.x;  // opdate the old locations
	OldY = MouseCoord.y;  // opdate the old locations
}

//*=================================================================================*
// CleanUp - cleans up and dynamic memory used
//*=================================================================================*
void CleanUp()
{
	delete VirtualScreen;

	FreeIntensityTable();
	FreeTransparencyTable();
	FreeTexture(0);
	FreeTexture(1);
	FreeVectorObject(&Obj);
}

//*=================================================================================*
// CreateVirtualPage - creates a virtual page with palette
//*=================================================================================*
int CreateVirtualPage(HWND hwnd)
{
	unsigned char TempPalette[256*3];
	int  i;

	for (i = 0; i < 256; i++)
	{
		TempPalette[i*3+0] = GlobalPalette[i].r << 2;
		TempPalette[i*3+1] = GlobalPalette[i].g << 2;
		TempPalette[i*3+2] = GlobalPalette[i].b << 2;
	}
		
	// create an 8-bit device independant bitmap.
	// The first parameter is the HWND of the window we are using
	// The second parameter is the rectangular drawing area. If NULL, the size of the window is used
	// The third parameter is the palette we want to use for the window

	VirtualScreen = new CDIBSection8(hwnd, NULL, TempPalette);

	if (VirtualScreen == NULL)
	{
		MessageBox(NULL, "Error! Virtual page could not be allocated.", ErrorCaption, MB_OK);
		return(0);
	}

	// Initialize the buffer to all black
	memset(VirtualScreen->GetBuffer(), 0, VirtualScreen->GetPitch() * VirtualScreen->GetHeight());

	return (1);
}

//*=================================================================================*
// GeneralInit - Performs Inertia initialization
//*=================================================================================*
int GeneralInit()
{
	InitInertia();   // initialize Inertia
	
	// Loat the intensity table
	if (LoadIntensityTable("COPPRENV.IT", 0) != I_OK) 
	{	
     	MessageBox(NULL, "Error! Intensity table data file could not be loaded.", ErrorCaption, MB_OK);
		return (0);
	}

	// Load the transparency Table
	if (LoadTransparencyTable("COPPRENV.TT", 0) != I_OK) 
	{	
     	MessageBox(NULL, "Error! Transparency table data file could not be loaded.", ErrorCaption, MB_OK);
		return (0);
	}

	// Load the texture map
	if (LoadPCXTexture(1, "COPPRENV.PCX", 0) != I_OK) 
	{	
     	MessageBox(NULL, "Error! Texture map could not be loaded.", ErrorCaption, MB_OK);
		return (0);
	}

	// create a phong highlight map
	if (MakePhongMap(0) != I_OK)
	{
		MessageBox(NULL, "Error! Phong intensity map could not be created.", ErrorCaption, MB_OK);
		return(0);
	}

	InitView(&View);

	GetTexturePalette(1);  // grab the palette from the first texture	

	PushLightSource(0, 0, -1);   // add a light source to the scene

	return(1);  // if we make it this far, all is good, so return success
}